// Dutchtronix AVR Oscilloscope Clock
//
//  Copyright  2010-2021 Johannes P.M. de Rie
//
//  All Rights Reserved
//
//  This file is part of the Dutchtronix Oscilloscope Clock Distribution.
//  Use, modification, and re-distribution is permitted subject to the
//  terms in the file named "LICENSE.TXT", which contains the full text
//  of the legal notices and should always accompany this Distribution.
//
//  This software is provided "AS IS" with NO WARRANTY OF ANY KIND.
//
//  This notice (including the copyright and warranty disclaimer)
//  must be included in all copies or derivations of this software.
//
// Firmware for Dutchtronix AVR Oscilloscope Clock
//
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "./ClkConfig.h"
#include "./ClkData.h"
#include "./ClkISR.h"

//
// Output frequency (using 24 bit accumulator) :
//
//	f = deltaPhase * fClock/2^24
//
//   fClock is in this case the CPU clock divided by the
//	number of cycles to output the data (12 cycles ->1666666.6)
//
//	f = ADDER0/ADDER1/ADDER2 * (1666666.6/16777216)
//
//	f = ADDER0/ADDER1/ADDER2 * 0.099341
//
//	fMax (theoretical) = 0.5 * fClock
//	Observed decent frequency:  KHz (adder value 0x)
//
// computed as 1000 / .099341
#define OneKHZ 	10066

extern void SpecialWait4UART(void);
extern void Special_UARTPfu08(byte b);
extern void UsrSendByte(char ch);
extern const byte SineTbl[] PROGMEM;

extern void up_one(void);
extern void up_hundred(void);
extern void up_tenthousand(void);
extern void down_one(void);
extern void down_hundred(void);
extern void down_tenthousand(void);
extern unsigned long GetAdder(void);
extern void SetAdder(unsigned long l);

void StartMiniDDS(void);
void StopMiniDDS(void);
pByte CopyWaveTable(PGM_P pSrc);
void SendAdderValue(void);
void ReadAdderValue(void);

void StartMiniDDS(void)
{
	SpecialWait4UART();				//wait for I/O to complete
	cli();
	DisableINT0();					//Stop INT0 interrupts
	DisableTC0();					//Stop TC0 Timer Interrupts
	DisableTC1();					//Stop TC1 Timer Interrupts	Check may not be running
	DisableTC2();					//Stop TC2 Timer Interrupts
#if !ASCIILINK
	UARTSetBaudVal(4);				//DDSControl.exe expects 115,200
#endif
	fMiniDDS = TRUE;				//Set MiniDDS mode
	sei();
}
//
// Stopping MiniDDS is not trivial since we don't want to poll
//
// Suggested solution: From the serial input function, when an exit
//	command is detected, PATCH the code in wavegen to stop the loop
//	and execute StopMiniDDS
//
void StopMiniDDS(void)
{
	InitiateSysReset();
}
pByte SetWaveTable(char ch)
{
//
// Create the various Waveform Tables in the SRAM buffer
// All but SineTbl are generated on the fly.
// ch is expected to be in the range '1' to '4'. If not,
// the buffer is set to 0xfe.
//
	pByte pDst = WaveTbl;			//destination
	byte cnt = 0, v;				//default copy 256 bytes
	pByte pSrc = (pByte)SineTbl;	//for Sine
	v = 0xfe;						//for Triangle
	do {
		if (ch == '1') {			//Sine Table
			v = pgm_read_byte(pSrc++);
		} else if (ch == '2') {		//Sawtooth Table
			v = cnt;
		} else if (ch == '3') {		//Triangle Table
			if (cnt < 128) {
				v += 2;
			} else {
				if (cnt == 128) v = 0x01;
				v -= 2;
			}
		} else if (ch == '4') {		//Square Table
			if (cnt < 128) {
				v = 0;
			} else {
				v = 0xff;
			}
		}
		*pDst++ = v;
	} while (++cnt);				//until overflow
	return WaveTbl;
}
//	
// send the current adder value (frequency) to the PC
// as a 5 byte sequence :
// 'F' folowed by a 32 bit phase accumulator value
//	  or 4 hex values in ASCIILINK
//
void SendAdderValue(void)
{
	union {
		unsigned long l;
		unsigned char b[4];					//Little Endian
	} adder;
	adder.l = GetAdder();

	_delay_ms(10);
#if ASCIILINK
	UsrSendByte('F');						// flag
	Special_UARTPfu08(0);					// zero byte for 32-bit compatibility
	Special_UARTPfu08(adder.b[2]);			// high adder
	Special_UARTPfu08(adder.b[1]);			// mid adder
	Special_UARTPfu08(adder.b[0]);			// low adder
	UsrSendByte(CR);						// terminator
	UsrSendByte(LF);						// terminator
#else
	UsrSendByte('F');						// flag
	UsrSendByte(0);							// zero byte for 32-bit compatibility
	UsrSendByte(adder.b[2]);				// high adder
	UsrSendByte(adder.b[1]);				// mid adder
	UsrSendByte(adder.b[0]);				// low adder
	UsrSendByte(LF);						// terminator
#endif
}


#if ASCIILINK
byte getnibble(void)
{
	char ch = UARTReceiveByte();
	if (ch < '0') {
		return 0;
	}
	ch -= '0';
	if (ch > 9) {
		ch -= 7;
	}
	return ch;
}
//
// read 6 hex characters (24-bit value) from the serial link
// no error checking
//
void ReadAdderValue(void)
{
	union {
		unsigned long l;
		unsigned char b[4];				//Little Endian
	} adder;
	adder.b[3] = 0;
	adder.b[2] = (getnibble() << 4) | getnibble();	//MSB
	adder.b[1] = (getnibble() << 4) | getnibble();	//Middle SB
	adder.b[0] = (getnibble() << 4) | getnibble();	//LSB
	SetAdder(adder.l);
}	


#else
// 
// read in 4 bytes from the serial link
//
void ReadAdderValue(void)
{
	union {
		unsigned long l;
		unsigned char b[4];				//Little Endian
	} adder;

	(void)UARTReceiveByte();			// read and ignore bits 32..24
	adder.b[3] = 0;
	adder.b[2] = UARTReceiveByte();		// read bits 23..16: MSB
	adder.b[1] = UARTReceiveByte();		// read bits 15..8: MIDDLE SB
	adder.b[0] = UARTReceiveByte();		// read bits 7..0: LSB
	SetAdder(adder.l);
}
#endif
//
// Called from Timer1 ISR, therefore all "Call Used" registers have been saved.
//
void FuncGenProcessInput(void)
{
	DisableTC2();						//no need for TC2 anymore
	Flags.InputActive = TRUE;			//stop enabling TC1 when input arrives
//
// use UARTReceiveByte() to get serial input.
// use UsrSendByte(ch) to send serial output.
// These calls may block
//
//
// Check if the push button switch S1/S2 is closed. Reboot if it is
//
#if ONESWITCH
#if PORTDSWITCH
	SWPort |= (_BV(SW1Bit));			//re-enable pull-up resistor since CharPixel() will undo it.
	asm("nop");
#endif

#else
	if ((SWPin & _BV(SW2Bit)) == 0)	{	//Low means push button pressed
		InitiateSysReset();
	}
#endif
	sei();								//allow more serial input
	byte ch = UARTReceiveByte();
#if ONESWITCH
	//workaround!
	if ((SWPin & _BV(SW1Bit)) == 0)	{	//Low means push button pressed
		InitiateSysReset();
	}
#endif
#if ASCIILINK
	UsrSendByte(ch);
#endif
	// DDSControl.exe (Jesper Hansen) protocol
	if (ch == '-') {
		up_one();
	} else if (ch == 'd') {
		up_hundred();
	} else if (ch == 'D') {
		up_tenthousand();
	} else if (ch == '+') {
		down_one();
	} else if (ch == 'u') {
		down_hundred();
	} else if (ch == 'U') {
		down_tenthousand();
	} else if (ch == 's') {
		ReadAdderValue();
#if ASCIILINK
	} else if (ch == 'q') {
		StopMiniDDS();
#endif
	} else {
		(void)SetWaveTable(ch);
	}
//
// always reply with the current adder value
//
	SendAdderValue();
	while (!IsUARTOutReady()) ;			//wait for send to complete
	cli();
	Flags.InputActive =  FALSE;			//start enabling TC2 when input arrives
}

const byte SineTbl[256] PROGMEM = {			// 256 step sinewave table
		0x80,0x83,0x86,0x89,0x8c,0x8f,0x92,0x95,0x98,0x9c,0x9f,0xa2,0xa5,0xa8,0xab,0xae,
		0xb0,0xb3,0xb6,0xb9,0xbc,0xbf,0xc1,0xc4,0xc7,0xc9,0xcc,0xce,0xd1,0xd3,0xd5,0xd8,
		0xda,0xdc,0xde,0xe0,0xe2,0xe4,0xe6,0xe8,0xea,0xec,0xed,0xef,0xf0,0xf2,0xf3,0xf5,
		0xf6,0xf7,0xf8,0xf9,0xfa,0xfb,0xfc,0xfc,0xfd,0xfe,0xfe,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xfe,0xfe,0xfd,0xfc,0xfc,0xfb,0xfa,0xf9,0xf8,0xf7,
		0xf6,0xf5,0xf3,0xf2,0xf0,0xef,0xed,0xec,0xea,0xe8,0xe6,0xe4,0xe2,0xe0,0xde,0xdc,
		0xda,0xd8,0xd5,0xd3,0xd1,0xce,0xcc,0xc9,0xc7,0xc4,0xc1,0xbf,0xbc,0xb9,0xb6,0xb3,
		0xb0,0xae,0xab,0xa8,0xa5,0xa2,0x9f,0x9c,0x98,0x95,0x92,0x8f,0x8c,0x89,0x86,0x83,
		0x80,0x7c,0x79,0x76,0x73,0x70,0x6d,0x6a,0x67,0x63,0x60,0x5d,0x5a,0x57,0x54,0x51,
		0x4f,0x4c,0x49,0x46,0x43,0x40,0x3e,0x3b,0x38,0x36,0x33,0x31,0x2e,0x2c,0x2a,0x27,
		0x25,0x23,0x21,0x1f,0x1d,0x1b,0x19,0x17,0x15,0x13,0x12,0x10,0x0f,0x0d,0x0c,0x0a,
		0x09,0x08,0x07,0x06,0x05,0x04,0x03,0x03,0x02,0x01,0x01,0x00,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x02,0x03,0x03,0x04,0x05,0x06,0x07,0x08,
		0x09,0x0a,0x0c,0x0d,0x0f,0x10,0x12,0x13,0x15,0x17,0x19,0x1b,0x1d,0x1f,0x21,0x23,
		0x25,0x27,0x2a,0x2c,0x2e,0x31,0x33,0x36,0x38,0x3b,0x3e,0x40,0x43,0x46,0x49,0x4c,
		0x4f,0x51,0x54,0x57,0x5a,0x5d,0x60,0x63,0x67,0x6a,0x6d,0x70,0x73,0x76,0x79,0x7c
};

